//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Detector/DetectorItem.cpp
//! @brief     Implements classes DetectorItems.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Detector/DetectorItem.h"
#include "Base/Axis/Scale.h"
#include "Base/Const/Units.h"
#include "Base/Util/Assert.h"
#include "Device/Detector/Detector2D.h"
#include "Device/Mask/IShape2D.h"
#include "Device/Resolution/ResolutionFunction2DGaussian.h"
#include "GUI/Model/Detector/ResolutionFunctionItems.h"
#include "GUI/Model/Mask/MaskContainerItem.h"
#include "GUI/Model/Mask/MaskItems.h"
#include "GUI/Support/XML/Backup.h"

namespace {
namespace Tag {

const QString ExpandResolutionFuncGroupbox("ExpandResolutionFuncGroupbox");
const QString AlphaAxis("AlphaAxis");
const QString PhiAxis("PhiAxis");
const QString ResolutionFunction("ResolutionFunction");

} // namespace Tag

} // namespace


DetectorItem::DetectorItem()
{
    m_phiAxis.initWidth("Span (deg)", "Full width in phi direction", 4.0,
                        RealLimits::limited(0, 180));
    m_phiAxis.initCenter("Center (deg)", "Angle phi of detector center", 0.,
                         RealLimits::limited(-90, 90));

    m_alphaAxis.initWidth("Span (deg)", "Full width in alpha direction", 4.0,
                          RealLimits::limited(0, 180));
    m_alphaAxis.initCenter("Center (deg)", "Angle alpha of detector center", 0.,
                           RealLimits::limited(-90, 90));

    m_resolutionFunction.init("Resolution function", "Detector resolution function",
                              ResolutionFunctionItemCatalog::Type::None);
}

DetectorItem::~DetectorItem() = default;

std::unique_ptr<IDetector> DetectorItem::createDetector() const
{
    const int n_x = m_phiAxis.nbins();
    const double x_wid = Units::deg2rad(m_phiAxis.width());
    const double x_cen = Units::deg2rad(m_phiAxis.center());

    const int n_y = m_alphaAxis.nbins();
    const double y_wid = Units::deg2rad(m_alphaAxis.width());
    const double y_cen = Units::deg2rad(m_alphaAxis.center());

    auto result = std::make_unique<Detector2D>(x_wid, y_wid, n_x, n_y, x_cen, y_cen);

    if (m_maskContainerItem) {
        const QVector<MaskItem*> maskItems = m_maskContainerItem->maskItems();
        for (auto maskIter = maskItems.rbegin(); maskIter != maskItems.rend(); maskIter++) {
            if ((*maskIter)->isVisible()) {
                if (auto* ii = dynamic_cast<RegionOfInterestItem*>(*maskIter)) {
                    result->setRegionOfInterest(ii->xLow(), ii->yLow(), ii->xUp(), ii->yUp());
                } else {
                    std::unique_ptr<IShape2D> shape((*maskIter)->createShape());
                    result->addBinMask(*shape, (*maskIter)->maskValue());
                }
            }
        }
    }

    if (auto resFunc = createResolutionFunction())
        result->setResolutionFunction(*resFunc);

    return result;
}

void DetectorItem::importMasks(const MaskContainerItem* item)
{
    ASSERT(item);
    m_maskContainerItem = std::make_unique<MaskContainerItem>();
    GUI::Util::copyContents(item, m_maskContainerItem.get());
}

void DetectorItem::setMasks(MaskContainerItem* item)
{
    m_maskContainerItem.reset(item);
}

SelectionProperty<ResolutionFunctionItemCatalog>& DetectorItem::resolutionFunctionSelection()
{
    return m_resolutionFunction;
}

std::unique_ptr<IResolutionFunction2D> DetectorItem::createResolutionFunction() const
{
    return m_resolutionFunction->createResolutionFunction();
}

void DetectorItem::writeTo(QXmlStreamWriter* w) const
{
    XML::writeAttribute(w, XML::Attrib::version, uint(2));

    // resolution function
    w->writeStartElement(Tag::ResolutionFunction);
    m_resolutionFunction.writeTo(w);
    w->writeEndElement();

    // resolution function groupbox: is expanded?
    w->writeStartElement(Tag::ExpandResolutionFuncGroupbox);
    XML::writeAttribute(w, XML::Attrib::value, expandResolutionFunc);
    w->writeEndElement();

    // phi axis
    w->writeStartElement(Tag::PhiAxis);
    m_phiAxis.writeTo(w);
    w->writeEndElement();

    // alpha axis
    w->writeStartElement(Tag::AlphaAxis);
    m_alphaAxis.writeTo(w);
    w->writeEndElement();
}

void DetectorItem::readFrom(QXmlStreamReader* r)
{
    const uint version = XML::readUIntAttribute(r, XML::Attrib::version);
    Q_UNUSED(version)

    while (r->readNextStartElement()) {
        QString tag = r->name().toString();

        // resolution function
        if (tag == Tag::ResolutionFunction) {
            m_resolutionFunction.readFrom(r);
            XML::gotoEndElementOfTag(r, tag);

            // resolution function groupbox: is expanded?
        } else if (tag == Tag::ExpandResolutionFuncGroupbox) {
            XML::readAttribute(r, XML::Attrib::value, &expandResolutionFunc);
            XML::gotoEndElementOfTag(r, tag);

            // x size

            // phi axis
        } else if (tag == Tag::PhiAxis) {
            m_phiAxis.readFrom(r);
            XML::gotoEndElementOfTag(r, tag);

            // alpha axis
        } else if (tag == Tag::AlphaAxis) {
            m_alphaAxis.readFrom(r);
            XML::gotoEndElementOfTag(r, tag);

        } else
            r->skipCurrentElement();
    }
}
