//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Device/DatafileItem.h
//! @brief     Defines class DatafileItem.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_GUI_MODEL_DEVICE_DATAFILEITEM_H
#define BORNAGAIN_GUI_MODEL_DEVICE_DATAFILEITEM_H

#include <QObject>
#include <QXmlStreamReader>
#include <QXmlStreamWriter>

class Data1DItem;
class Data2DItem;
class DataItem;
class Datafield;
class InstrumentItem;
class MessageService;

//! Provides access to experimental data, for display and fitting.

class DatafileItem : public QObject {
    Q_OBJECT
public:
    DatafileItem();
    ~DatafileItem();

    //! The name which is presented to the user
    QString realItemName() const { return m_name; }
    void setDatafileItemName(const QString& name);

    QString presentationType() const { return m_presentationType; }
    void setPresentationType(const QString& type);

    // data

    DataItem* dataItem() const;
    Data2DItem* data2DItem() const;
    Data1DItem* data1DItem() const;

    void initItemWithRank(int rank);

    size_t rank() const;

    bool holdsDimensionalData() const;
    void setTheDatafield(const Datafield& data);

    // instrument

    QString instrumentId() const { return m_instrumentId; }
    void setInstrumentId(const QString& id);
    void linkToInstrument(const InstrumentItem* instrument);
    void unlinkFromInstrument();

    //! Returns true, if rotation will affect linked instrument or mask presence.
    bool rotationAffectsSetup() const;
    void rotateData();

    void updateFileName();

    // write/read

    void writeTo(QXmlStreamWriter* w) const;
    void readFrom(QXmlStreamReader* r);

    void writeDatafiles(const QString& projectDir) const;
    QString readDatafiles(const QString& projectDir, MessageService* messageService);

    // other

    //! Make a copy instance.
    void copyTo(DatafileItem* const dest) const;

    //! Returns the shape of underlying data item
    std::vector<int> shape() const;

signals:
    void importContentsProcessed();

private:
    //! This does not set the link to the instrument! Use linkToInstrument() if you want
    //! to link to this instrument.
    void updateToInstrument(const InstrumentItem* instrument);

    QString m_instrumentId;
    QString m_name;
    QString m_presentationType;

    // DataItem can be `Data2DItem` and `Data1DItem` (default `Data2DItem`).
    std::unique_ptr<DataItem> m_dataItem;
};

#endif // BORNAGAIN_GUI_MODEL_DEVICE_DATAFILEITEM_H
