//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Combo/ItemComboWidget.cpp
//! @brief     Implements class ItemComboWidget.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Combo/ItemComboWidget.h"
#include "Base/Util/Assert.h"
#include "GUI/Model/Device/DatafileItem.h"
#include "GUI/Model/Job/JobItem.h"
#include "GUI/View/Access/DataAccessWidget.h"
#include "GUI/View/Combo/ItemComboToolbar.h"
#include <QVBoxLayout>

ItemComboWidget::ItemComboWidget()
    : m_toolbar(new ItemComboToolbar)
    , m_stackedWidget(new QStackedWidget)
    , m_blankWidget(new QWidget)
{
    setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding);
    m_stackedWidget->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding);

    m_stackedWidget->addWidget(m_blankWidget);

    auto* layout = new QVBoxLayout;
    layout->setContentsMargins(0, 0, 0, 0);
    layout->setSpacing(0);
    layout->addWidget(m_toolbar);
    layout->addWidget(m_stackedWidget);
    setLayout(layout);

    connect(m_toolbar, &ItemComboToolbar::comboChanged, this, &ItemComboWidget::onComboChanged);

    setToolbarVisible(false);
}

void ItemComboWidget::registerWidget(const QString& type, factory_function_t f)
{
    m_widgetFactory.registerItem(type, f);
}

//! Sets stack to show widget corresponding to given presentation
void ItemComboWidget::setComboPresentation(const QString& type)
{
    if (type.isEmpty()) {
        m_stackedWidget->setCurrentWidget(m_blankWidget);
        setToolbarVisible(false);
        return;
    }

    if (!activePresentationList().contains(type))
        return;

    setToolbarVisible(true);
    m_toolbar->setPresentation(type);

    DataAccessWidget* widget = m_typeToWidget[type];

    if (!widget) {
        widget = m_widgetFactory.createItemPtr(type).release();
        m_stackedWidget->addWidget(widget);
        m_typeToWidget[type] = widget;
    }
    ASSERT(widget);
    m_toolbar->setActionList(widget->actionList());
    m_stackedWidget->setCurrentWidget(widget);
    if (widget->isHidden())
        widget->show();

    setSizeToCurrentWidget();

    ASSERT(m_item);
    widget->setJobOrDatafileItem(m_item);
}

void ItemComboWidget::setToolbarVisible(bool value)
{
    m_toolbar->setVisible(value);
}

void ItemComboWidget::setTheItem(QObject* item)
{
    if (!item) {
        m_stackedWidget->setCurrentWidget(m_blankWidget);
        return;
    }

    m_item = item;
    ASSERT(jobItem() || realItem());

    disconnect(m_item, &QObject::destroyed, this, nullptr);
    connect(m_item, &QObject::destroyed, this, [this]() { m_item = nullptr; });

    setToolbarVisible(!itemPresentation().isEmpty());
    m_toolbar->setPresentationList(presentationList(), activePresentationList());
    setThePresentation(itemPresentation());
}

//! Returns full list of presentations available for given item.

QStringList ItemComboWidget::presentationList() const
{
    return activePresentationList();
}

//! Presentation which should be shown for current item.

QString ItemComboWidget::defaultPresentation() const
{
    if (!activePresentationList().empty())
        return activePresentationList().first();

    return selectedPresentation();
}

//! Presentation selected in combo selector.

QString ItemComboWidget::selectedPresentation() const
{
    return m_toolbar->currentPresentation();
}

void ItemComboWidget::onComboChanged(const QString&)
{
    setThePresentation(selectedPresentation());
}

//! Resizes QStackedWidget to currently active page.

void ItemComboWidget::setSizeToCurrentWidget()
{
    for (int i = 0; i < m_stackedWidget->count(); ++i) {
        // determine the vertical size policy
        QSizePolicy::Policy policy = QSizePolicy::Ignored;
        if (i == m_stackedWidget->currentIndex())
            policy = QSizePolicy::Expanding;

        // update the size policy
        auto* page = m_stackedWidget->widget(i);
        page->setSizePolicy(policy, policy);
    }
}

JobItem* ItemComboWidget::jobItem() const
{
    return dynamic_cast<JobItem*>(m_item);
}

DatafileItem* ItemComboWidget::realItem() const
{
    return dynamic_cast<DatafileItem*>(m_item);
}
