//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Frame/MaskEditingFrame.cpp
//! @brief     Implements class MaskEditingFrame.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Frame/MaskEditingFrame.h"
#include "GUI/Model/Data/Data2DItem.h"
#include "GUI/Model/Mask/MaskContainerModel.h"
#include "GUI/Model/Mask/MaskItems.h"
#include "GUI/View/Canvas/MaskEditorActions.h"
#include "GUI/View/Canvas/MaskEditorCanvas.h"
#include "GUI/View/Canvas/MaskEditorToolbar.h"
#include "GUI/View/Scene/MaskGraphicsScene.h"
#include "GUI/View/Setup/MaskPanel.h"
#include <QSplitter>
#include <QVBoxLayout>

MaskEditingFrame::MaskEditingFrame(QWidget* parent)
    : DataAccessWidget(parent)
    , m_actions(new MaskEditorActions(this))
    , m_toolpanel(new MaskPanel)
    , m_canvas(new MaskEditorCanvas)
{
    setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding);

    auto* layout = new QHBoxLayout;
    setLayout(layout);
    layout->setContentsMargins(0, 0, 0, 0);
    layout->setSpacing(0);

    auto* splitter = new QSplitter;
    layout->addWidget(splitter);
    splitter->addWidget(m_canvas);
    splitter->addWidget(m_toolpanel);
    splitter->setCollapsible(1, true);

    auto* toolbar = new MaskEditorToolbar(m_actions);
    layout->addWidget(toolbar);
    toolbar->setOrientation(Qt::Vertical);
    toolbar->setStyleSheet("QToolBar{border-left:1px solid rgb(180,180,180);}");

    m_toolpanel->setPanelHidden(true);

    //... Connections

    // reset view request is propagated from editorActions to graphics view
    connect(m_actions, &MaskEditorActions::resetViewRequest, m_canvas,
            &MaskEditorCanvas::onResetViewRequest);

    // tool panel request is propagated from editorActions to this MaskEditor
    connect(m_actions, &MaskEditorActions::propertyPanelRequest, this,
            &MaskEditingFrame::onPropertyPanelRequest);

    // save plot request is propagated from editorActions to graphics scene
    connect(m_actions, &MaskEditorActions::savePlotRequest, m_canvas,
            &MaskEditorCanvas::onSavePlotRequest);

    // selection/drawing activity is propagated from Toolbar to graphics scene
    connect(toolbar, &MaskEditorToolbar::activityChanged, m_canvas->getScene(),
            &MaskGraphicsScene::onActivityChanged);

    // mask value is propagated from Toolbar to graphics scene
    connect(toolbar, &MaskEditorToolbar::maskValueChanged, m_canvas->getScene(),
            &MaskGraphicsScene::onMaskValueChanged);

    // show results request is propagated from Toolbar to Canvas
    connect(toolbar, &MaskEditorToolbar::presentationChange, m_canvas,
            &MaskEditorCanvas::onPresentationChange);

    // space bar push (request for zoom mode) is propagated from graphics view to Toolbar
    connect(m_canvas, &MaskEditorCanvas::changeActivityRequest, toolbar,
            &MaskEditorToolbar::onChangeActivityRequest);

    // Delete request is propagated from canvas to actions
    connect(m_canvas, &MaskEditorCanvas::deleteSelectedRequest, m_actions,
            &MaskEditorActions::onDeleteMaskAction);

    // context menu request is propagated from graphics scene to MaskEditorActions
    connect(m_canvas->getScene(), &MaskGraphicsScene::itemContextMenuRequest, m_actions,
            &MaskEditorActions::onItemContextMenuRequest);

    // context menu request is propagated from PropertyPanel to MaskEditorActions
    connect(m_toolpanel, &MaskPanel::itemContextMenuRequest, m_actions,
            &MaskEditorActions::onItemContextMenuRequest);
}

void MaskEditingFrame::setJobOrDatafileItem(QObject* item)
{
    setBaseItem(item);
    setContextMW();
}

//! shows/hides right panel with properties
void MaskEditingFrame::onPropertyPanelRequest()
{
    m_toolpanel->setPanelHidden(!m_toolpanel->isHidden());
}

//! Returns list of actions intended for styled toolbar (on the top).

QList<QAction*> MaskEditingFrame::actionList()
{
    return m_actions->topToolbarActions();
}

void MaskEditingFrame::showEvent(QShowEvent*)
{
    if (currentData2DItem())
        setContextMW();
}

void MaskEditingFrame::hideEvent(QHideEvent*)
{
    m_toolpanel->resetContextMPP();
    m_canvas->resetContextMEC();
}

void MaskEditingFrame::setContextMW()
{
    ASSERT(currentData2DItem());

    MaskContainerModel* containerModel = currentData2DItem()->getOrCreateMaskContainerModel();
    ASSERT(containerModel);

    m_toolpanel->setContextMPP(containerModel);

    m_canvas->setContextMEC(currentData2DItem(), m_toolpanel->selectionModel());

    m_actions->setModels(containerModel, m_toolpanel->selectionModel());

    update();
}
