//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Shape/IShapeDisplay.cpp
//! @brief     Implements interface IShapeDisplay.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Shape/IShapeDisplay.h"
#include "GUI/Model/Mask/MaskItems.h"
#include "GUI/View/ProShape/MaskUtil.h"
#include "GUI/View/ProShape/SceneAdaptor.h"
#include <QAction>
#include <QGraphicsScene>
#include <QGraphicsSceneContextMenuEvent>
#include <QMenu>

namespace {

bool itemMaskValue(const MaskObject* item)
{
    if (const auto* maskItem = dynamic_cast<const MaskItem*>(item))
        return maskItem->maskValue();
    return false;
}

} // namespace

IShapeDisplay::IShapeDisplay(MaskObject* item)
    : m_adaptor(nullptr)
    , m_block_on_property_change(false)
{
    ASSERT(item);

    connect(this, &IShapeDisplay::xChanged, this, &IShapeDisplay::onChangedX, Qt::UniqueConnection);
    connect(this, &IShapeDisplay::yChanged, this, &IShapeDisplay::onChangedY, Qt::UniqueConnection);

    connect(item, &MaskObject::maskGeometryChanged, this, &IShapeDisplay::onGeometryChange,
            Qt::UniqueConnection);
    connect(item, &MaskObject::maskVisibilityChanged, this, &IShapeDisplay::onVisibilityChange,
            Qt::UniqueConnection);
}

QRectF IShapeDisplay::boundingRect() const
{
    return m_bounding_rect;
}

void IShapeDisplay::setSceneAdaptor(const SceneAdaptor* adaptor)
{
    ASSERT(adaptor);
    if (m_adaptor == adaptor)
        return;

    if (m_adaptor)
        disconnect(m_adaptor, &SceneAdaptor::update_request, this, &IShapeDisplay::update_view);

    m_adaptor = adaptor;
    connect(m_adaptor, &SceneAdaptor::update_request, this, &IShapeDisplay::update_view,
            Qt::UniqueConnection);
    update_view();
}

void IShapeDisplay::paint(QPainter* painter, const QStyleOptionGraphicsItem*, QWidget*)
{
    const MaskObject* obj = parameterizedItem();
    ASSERT(obj);
    const bool is_masked = itemMaskValue(obj);
    painter->setBrush(MaskUtil::getMaskBrush(is_masked));
    painter->setPen(MaskUtil::getMaskPen(is_masked));
    painter->setRenderHints(QPainter::Antialiasing);
    painter->drawPath(maskedShape());
}

QPainterPath IShapeDisplay::maskedShape() const
{
    static const QSet<MaskType> relevantMaskType = {MaskType::RECTANGLE, MaskType::POLYGON,
                                                    MaskType::VERTICALLINE,
                                                    MaskType::HORIZONTALLINE, MaskType::ELLIPSE};

    QPainterPath result = mapToScene(shape());
    for (const QGraphicsItem* item : scene()->items()) {
        const auto* view = dynamic_cast<const IShapeDisplay*>(item);
        if (!view)
            continue;
        if (!relevantMaskType.contains((MaskType)view->type()))
            continue;
        if (itemMaskValue(view->parameterizedItem()))
            continue;
        if (const auto* mask_item = dynamic_cast<const MaskItem*>(view->parameterizedItem()))
            if (!mask_item->isVisible())
                continue;
        if (!(view->zValue() > zValue()))
            continue;

        const QPainterPath mask_shape = view->mapToScene(view->shape());
        if (!mask_shape.intersects(result))
            continue;

        // Item lays on top and is non-masking -> subtract the path of the item
        result = result.subtracted(mask_shape);
    }

    return mapFromScene(result);
}

qreal IShapeDisplay::toSceneX(qreal value) const
{
    return m_adaptor ? m_adaptor->toSceneX(value) : value;
}

qreal IShapeDisplay::toSceneY(qreal value) const
{
    return m_adaptor ? m_adaptor->toSceneY(value) : value;
}

qreal IShapeDisplay::fromSceneX(qreal value) const
{
    return m_adaptor ? m_adaptor->fromSceneX(value) : value;
}

qreal IShapeDisplay::fromSceneY(qreal value) const
{
    return m_adaptor ? m_adaptor->fromSceneY(value) : value;
}

void IShapeDisplay::addView(IShapeDisplay* childView)
{
    if (!childItems().contains(childView))
        childView->setParentItem(this);
}

void IShapeDisplay::setBlockOnProperty(bool value)
{
    m_block_on_property_change = value;
}

void IShapeDisplay::onGeometryChange()
{
    if (m_block_on_property_change)
        return;

    m_block_on_property_change = true;
    onPropertyChange();
    m_block_on_property_change = false;
}

void IShapeDisplay::onVisibilityChange()
{
    if (m_block_on_property_change)
        return;

    m_block_on_property_change = true;

    if (auto* maskItem = dynamic_cast<MaskItem*>(parameterizedItem()))
        this->setVisible(maskItem->isVisible());
    else
        this->setVisible(false);

    onPropertyChange();
    update();

    m_block_on_property_change = false;
}
