//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Shape/RectangleDisplay.cpp
//! @brief     Implements class RectangleDisplay.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Shape/RectangleDisplay.h"
#include "GUI/Model/Mask/MaskItems.h"
#include <QGraphicsSceneMouseEvent>
#include <QPainter>

RectangleDisplay::RectangleDisplay(RectangleItem* item)
    : RectangleBaseDisplay(item)
    , m_item(item)
{
    m_item->maskVisibilityChanged();
}

QPainterPath RectangleDisplay::shape() const
{
    QPainterPath path;
    path.addRect(m_mask_rect);
    return path;
}

MaskObject* RectangleDisplay::parameterizedItem() const
{
    return m_item;
}

void RectangleDisplay::onChangedX()
{
    setBlockOnProperty(true);
    m_item->setXLow(fromSceneX(this->x()));
    m_item->setXUp(fromSceneX(this->x() + m_mask_rect.width()));
    setBlockOnProperty(false);
}

void RectangleDisplay::onChangedY()
{
    setBlockOnProperty(true);
    m_item->setYLow(fromSceneY(this->y() + m_mask_rect.height()));
    m_item->setYUp(fromSceneY(this->y()));
    setBlockOnProperty(false);
}

void RectangleDisplay::onPropertyChange()
{
    update_view();
}

void RectangleDisplay::mouseMoveEvent(QGraphicsSceneMouseEvent* event)
{
    if (m_activeHandleElement) {
        qreal xmin = std::min(event->scenePos().x(), m_resize_opposite_origin.x());
        qreal xmax = std::max(event->scenePos().x(), m_resize_opposite_origin.x());
        qreal ymin = std::min(event->scenePos().y(), m_resize_opposite_origin.y());
        qreal ymax = std::max(event->scenePos().y(), m_resize_opposite_origin.y());

        if (m_activeHandleElement->getHandleType() == SizeHandleElement::RESIZE) {
            m_item->setXLow(fromSceneX(xmin));
            m_item->setYLow(fromSceneY(ymax));
            m_item->setXUp(fromSceneX(xmax));
            m_item->setYUp(fromSceneY(ymin));
        } else if (m_activeHandleElement->getHandleType() == SizeHandleElement::RESIZE_HEIGHT) {
            m_item->setYLow(fromSceneY(ymax));
            m_item->setYUp(fromSceneY(ymin));
        } else if (m_activeHandleElement->getHandleType() == SizeHandleElement::RESIZE_WIDTH) {
            m_item->setXLow(fromSceneX(xmin));
            m_item->setXUp(fromSceneX(xmax));
        }
        update_view();
    } else {
        QGraphicsObject::mouseMoveEvent(event);
    }
}

//! updates position of view using item properties
void RectangleDisplay::update_position()
{
    setX(toSceneX(m_item->xLow()));
    setY(toSceneY(m_item->yUp()));
}

QRectF RectangleDisplay::mask_rectangle()
{
    return {0, 0, width(), height()};
}

//! Returns the x-coordinate of the rectangle's left edge

qreal RectangleDisplay::left() const
{
    return toSceneX(m_item->xLow());
}

//! Returns the x-coordinate of the rectangle's right edge

qreal RectangleDisplay::right() const
{
    return toSceneX(m_item->xUp());
}

//! Returns the y-coordinate of the rectangle's top edge.

qreal RectangleDisplay::top() const
{
    return toSceneY(m_item->yUp());
}

//! Returns the y-coordinate of the rectangle's bottom edge.

qreal RectangleDisplay::bottom() const
{
    return toSceneY(m_item->yLow());
}
