#include "Base/Axis/MakeScale.h"

#include "Device/Data/Datafield.h"
#include "Device/IO/IOFactory.h"
#include "GUI/Model/Data/DataItem.h"
#include "GUI/Model/Device/BackgroundItems.h"
#include "GUI/Model/Device/DatafileItem.h"
#include "GUI/Model/Device/InstrumentItems.h"
#include "GUI/Model/Device/InstrumentModel.h"
#include "GUI/Model/Model/DatafilesModel.h"
#include "Tests/GTestWrapper/google_test.h"
#include "Tests/Unit/GUI/Utils.h"
#include <QTest>
#include <fstream>

const QString dir_test_real_model = "TestDatafilesModel";

TEST(TestDatafilesModel, saveNonXMLData)
{
    if (!QFile::exists(dir_test_real_model))
        QDir(".").mkdir(dir_test_real_model);

    const QString dir = dir_test_real_model + "/saveNonXMLData";
    UTest::GUI::create_dir(dir);

    const double value1(1.0), value2(2.0), value3(3.0);

    DatafilesModel model;
    DatafileItem* item1 = UTest::GUI::createRealData("data1", model, value1, 2);
    DatafileItem* item2 = UTest::GUI::createRealData("data2", model, value2, 2);

    item1->updateFileName();
    item2->updateFileName();

    // save first time
    model.writeDatafiles(dir);
    QTest::qSleep(10);

    // check existence of data on disk
    QString fname1 = dir + "/realdata_data1.int";
    QString fname2 = dir + "/realdata_data2.int";
    EXPECT_TRUE(QFile::exists(fname1));
    EXPECT_TRUE(QFile::exists(fname2));

    // read data from disk, checking it is the same
    std::unique_ptr<Datafield> dataOnDisk1(
        IO::readData2D(fname1.toStdString(), IO::Filetype2D::bornagain2D));
    std::unique_ptr<Datafield> dataOnDisk2(
        IO::readData2D(fname2.toStdString(), IO::Filetype2D::bornagain2D));
    EXPECT_TRUE(UTest::GUI::isTheSame(*dataOnDisk1, *item1->dataItem()->c_field()));
    EXPECT_TRUE(UTest::GUI::isTheSame(*dataOnDisk2, *item2->dataItem()->c_field()));

    // modify data and save the project
    item2->setTheDatafield(UTest::GUI::createData(value3, 2));
    model.writeDatafiles(dir);
    QTest::qSleep(10);

    EXPECT_TRUE(UTest::GUI::isTheSame(*dataOnDisk1, *item1->dataItem()->c_field()));
    EXPECT_FALSE(UTest::GUI::isTheSame(*dataOnDisk2, *item2->dataItem()->c_field()));

    // check that data on disk has changed
    dataOnDisk2.reset(IO::readData2D(fname2.toStdString(), IO::Filetype2D::bornagain2D));
    EXPECT_TRUE(UTest::GUI::isTheSame(*dataOnDisk2, *item2->dataItem()->c_field()));

    // rename RealData and check that file on disk changed the name
    item2->setDatafileItemName("data2new");
    item2->updateFileName();
    model.writeDatafiles(dir);
    QTest::qSleep(10);

    QString fname2new = "./" + dir + "/realdata_data2new.int";
    EXPECT_TRUE(QFile::exists(fname2new));
    EXPECT_TRUE(UTest::GUI::isTheSame(fname2new, *item2->dataItem()->c_field()));

    // check that file with old name was removed
    EXPECT_FALSE(QFile::exists(fname2));
}

TEST(TestDatafilesModel, saveXMLData)
{
    if (!QFile::exists(dir_test_real_model))
        QDir(".").mkdir(dir_test_real_model);

    const QString dir = dir_test_real_model + "/saveXMLData";
    UTest::GUI::create_dir(dir);

    InstrumentModel instrument_model;
    DatafilesModel model1;

    // add specular DatafileItems with non-default parameters
    auto* spec1 = model1.insertDataItem(1);
    spec1->setDatafileItemName("spec1");
    spec1->setPresentationType("Reflectometry (Graph only)");
    spec1->linkToInstrument(instrument_model.addInstrumentItem<SpecularInstrumentItem>());

    // add second specular DatafileItem
    auto* spec2 = model1.insertDataItem(1);
    spec2->setDatafileItemName("spec2");

    // add 2D DatafileItems with non-default parameters
    auto* intensity1 = model1.insertDataItem(2);
    auto* intensity2 = model1.insertDataItem(2);
    intensity1->setDatafileItemName("GISAS");
    intensity2->setDatafileItemName("OffSpec");
    intensity1->linkToInstrument(instrument_model.addInstrumentItem<GISASInstrumentItem>());
    intensity2->linkToInstrument(instrument_model.addInstrumentItem<OffspecInstrumentItem>());

    // set non-default top-level model parameters
    model1.setSelectedRank(2);
    model1.setSelectedIndex(1);

    // save data to project files
    const QString file1 = dir + "/Project_1.ba";
    const QString file2 = dir + "/Project_2.ba";

    // write changed model to disk
    const QString tag = "DatafilesModel";
    UTest::GUI::writeXMLFile<DatafilesModel>(file1, model1, tag);
    EXPECT_TRUE(QFile::exists(file1));

    // read data to the second model
    DatafilesModel model2;
    UTest::GUI::readXMLFile<DatafilesModel>(file1, model2, tag);

    // write the second model back to disk
    UTest::GUI::writeXMLFile<DatafilesModel>(file2, model2, tag);
    EXPECT_TRUE(QFile::exists(file2));

    // compare files byte-by-byte
    std::ifstream f1(file1.toStdString());
    std::ifstream f2(file2.toStdString());

    using it = std::istreambuf_iterator<char>;

    std::vector<char> contents1((it(f1)), it());
    std::vector<char> contents2((it(f2)), it());

    EXPECT_EQ(contents1, contents2);
}
